@@ -578,7 +578,6 @@ datafile_read(datafile_cur_t *cursor)
 Check to see if a file exists.
 Takes name of the file to check.
 @return true if file exists. */
-static
 bool
 file_exists(const char *filename)
 {
@@ -1540,13 +1539,14 @@ bool backup_start(CorruptedPages &corrupted_pages)
 		if (!write_galera_info(mysql_connection)) {
 			return(false);
 		}
-		write_current_binlog_file(mysql_connection);
 	}
 
-	if (opt_binlog_info == BINLOG_INFO_ON) {
+	bool with_binlogs = opt_binlog_info == BINLOG_INFO_ON;
 
-		lock_binlog_maybe(mysql_connection);
-		write_binlog_info(mysql_connection);
+	if (with_binlogs || opt_galera_info) {
+		if (!write_current_binlog_file(mysql_connection, with_binlogs)) {
+			return(false);
+		}
 	}
 
 	if (have_flush_engine_logs && !opt_no_lock) {
@@ -1580,15 +1580,34 @@ void backup_release()
 	}
 }
 
+static const char *default_buffer_pool_file = "ib_buffer_pool";
+
+static
+const char * get_buffer_pool_filename(size_t *length)
+{
+	/* If mariabackup is run for Galera, then the file
+	name is changed to the default so that the receiving
+	node can find this file and rename it according to its
+	settings, otherwise we keep the original file name: */
+	size_t dir_length = 0;
+	const char *dst_name = default_buffer_pool_file;
+	if (!opt_galera_info) {
+		dir_length = dirname_length(buffer_pool_filename);
+		dst_name = buffer_pool_filename + dir_length;
+	}
+	if (length) {
+		*length=dir_length;
+	}
+	return dst_name;
+}
+
 /** Finish after backup_start() and backup_release() */
 bool backup_finish()
 {
 	/* Copy buffer pool dump or LRU dump */
 	if (!opt_rsync) {
 		if (buffer_pool_filename && file_exists(buffer_pool_filename)) {
-			const char *dst_name;
-
-			dst_name = trim_dotslash(buffer_pool_filename);
+			const char *dst_name = get_buffer_pool_filename(NULL);
 			copy_file(ds_data, buffer_pool_filename, dst_name, 0);
 		}
 		if (file_exists("ib_lru_dump")) {
@@ -1677,17 +1696,14 @@ ibx_copy_incremental_over_full()
 
 		/* copy buffer pool dump */
 		if (innobase_buffer_pool_filename) {
-			const char *src_name;
-
-			src_name = trim_dotslash(innobase_buffer_pool_filename);
+			const char *src_name = get_buffer_pool_filename(NULL);
 
 			snprintf(path, sizeof(path), "%s/%s",
 				xtrabackup_incremental_dir,
 				src_name);
 
 			if (file_exists(path)) {
-				copy_file(ds_data, path,
-					  innobase_buffer_pool_filename, 0);
+				copy_file(ds_data, path, src_name, 0);
 			}
 		}
 
@@ -1922,6 +1938,14 @@ copy_back()
 
 	datadir_node_init(&node);
 
+	/* If mariabackup is run for Galera, then the file
+	name is changed to the default so that the receiving
+	node can find this file and rename it according to its
+	settings, otherwise we keep the original file name: */
+	size_t dir_length;
+	const char *src_buffer_pool;
+	src_buffer_pool = get_buffer_pool_filename(&dir_length);
+
 	while (datadir_iter_next(it, &node)) {
 		const char *ext_list[] = {"backup-my.cnf",
 			"xtrabackup_binary", "xtrabackup_binlog_info",
@@ -1984,6 +2008,11 @@ copy_back()
 			continue;
 		}
 
+		/* skip buffer pool dump */
+		if (!strcmp(filename, src_buffer_pool)) {
+			continue;
+		}
+
 		/* skip innodb data files */
 		is_ibdata_file = false;
 		for (Tablespace::const_iterator iter(srv_sys_space.begin()),
@@ -2006,23 +2035,18 @@ copy_back()
 
 	/* copy buffer pool dump */
 
-	if (innobase_buffer_pool_filename) {
-		const char *src_name;
-		char path[FN_REFLEN];
-
-		src_name = trim_dotslash(innobase_buffer_pool_filename);
-
-		snprintf(path, sizeof(path), "%s/%s",
-			mysql_data_home,
-			src_name);
-
-		/* could be already copied with other files
-		from data directory */
-		if (file_exists(src_name) &&
-			!file_exists(innobase_buffer_pool_filename)) {
-			copy_or_move_file(src_name,
-					  innobase_buffer_pool_filename,
-					  mysql_data_home, 0);
+	if (file_exists(src_buffer_pool)) {
+		char dst_dir[FN_REFLEN];
+		while (IS_TRAILING_SLASH(buffer_pool_filename, dir_length)) {
+			dir_length--;
+		}
+		memcpy(dst_dir, buffer_pool_filename, dir_length);
+		dst_dir[dir_length] = 0;
+		if (!(ret = copy_or_move_file(src_buffer_pool,
+				src_buffer_pool,
+				dst_dir, 1)))
+		{
+			goto cleanup;
 		}
 	}
 
