@@ -49,6 +49,7 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include <limits>
 #include "common.h"
 #include "xtrabackup.h"
+#include "srv0srv.h"
 #include "mysql_version.h"
 #include "backup_copy.h"
 #include "backup_mysql.h"
@@ -83,7 +84,6 @@ os_event_t	kill_query_thread_stop;
 bool sql_thread_started = false;
 char *mysql_slave_position = NULL;
 char *mysql_binlog_position = NULL;
-char *buffer_pool_filename = NULL;
 
 /* History on server */
 time_t history_start_time;
@@ -266,7 +266,8 @@ free_mysql_variables(mysql_variable *vars)
 
 static
 char *
-read_mysql_one_value(MYSQL *connection, const char *query)
+read_mysql_one_value(MYSQL *connection, const char *query,
+                     uint column, uint expect_columns)
 {
 	MYSQL_RES *mysql_result;
 	MYSQL_ROW row;
@@ -274,10 +275,10 @@ read_mysql_one_value(MYSQL *connection, const char *query)
 
 	mysql_result = xb_mysql_query(connection, query, true);
 
-	ut_ad(mysql_num_fields(mysql_result) == 1);
+	ut_ad(mysql_num_fields(mysql_result) == expect_columns);
 
 	if ((row = mysql_fetch_row(mysql_result))) {
-		result = strdup(row[0]);
+		result = strdup(row[column]);
 	}
 
 	mysql_free_result(mysql_result);
@@ -285,6 +286,15 @@ read_mysql_one_value(MYSQL *connection, const char *query)
 	return(result);
 }
 
+
+static
+char *
+read_mysql_one_value(MYSQL *mysql, const char *query)
+{
+  return read_mysql_one_value(mysql, query, 0/*offset*/, 1/*total columns*/);
+}
+
+
 static
 bool
 check_server_version(unsigned long version_number,
@@ -817,7 +827,7 @@ wait_for_no_updates(MYSQL *connection, uint timeout, uint threshold)
 
 static
 os_thread_ret_t
-kill_query_thread(
+DECLARE_THREAD(kill_query_thread)(
 /*===============*/
 	void *arg __attribute__((unused)))
 {
@@ -886,82 +896,6 @@ stop_query_killer()
 }
 
 
-/*
-Killing connections that wait for MDL lock.
-If lock-ddl-per-table is used, there can be some DDL statements
-
-FLUSH TABLES would hang infinitely, if DDL statements are waiting for
-MDL lock, which mariabackup currently holds. Therefore we start killing
-those  statements from a dedicated thread, until FLUSH TABLES WITH READ LOCK
-succeeds.
-*/
-
-static os_event_t mdl_killer_stop_event;
-static os_event_t mdl_killer_finished_event;
-
-static
-os_thread_ret_t
-DECLARE_THREAD(kill_mdl_waiters_thread(void *))
-{
-	MYSQL	*mysql;
-	if ((mysql = xb_mysql_connect()) == NULL) {
-		msg("Error: kill mdl waiters thread failed to connect");
-		goto stop_thread;
-	}
-
-	for(;;){
-		if (os_event_wait_time(mdl_killer_stop_event, 1000) == 0)
-			break;
-
-		MYSQL_RES *result = xb_mysql_query(mysql,
-			"SELECT ID, COMMAND, INFO FROM INFORMATION_SCHEMA.PROCESSLIST "
-			" WHERE State='Waiting for table metadata lock'",
-			true, true);
-		while (MYSQL_ROW row = mysql_fetch_row(result))
-		{
-			char query[64];
-
-			if (row[1] && !strcmp(row[1], "Killed"))
-				continue;
-
-			msg("Killing MDL waiting %s ('%s') on connection %s",
-				row[1], row[2], row[0]);
-			snprintf(query, sizeof(query), "KILL QUERY %s", row[0]);
-			if (mysql_query(mysql, query) && (mysql_errno(mysql) != ER_NO_SUCH_THREAD)) {
-				die("failed to execute query %s: %s", query,mysql_error(mysql));
-			}
-		}
-		mysql_free_result(result);
-	}
-
-	mysql_close(mysql);
-
-stop_thread:
-	msg("Kill mdl waiters thread stopped");
-	os_event_set(mdl_killer_finished_event);
-	os_thread_exit();
-	return os_thread_ret_t(0);
-}
-
-
-static void start_mdl_waiters_killer()
-{
-	mdl_killer_stop_event = os_event_create(0);
-	mdl_killer_finished_event = os_event_create(0);
-	os_thread_create(kill_mdl_waiters_thread, 0, 0);
-}
-
-
-/* Tell MDL killer to stop and finish for its completion*/
-static void stop_mdl_waiters_killer()
-{
-	os_event_set(mdl_killer_stop_event);
-	os_event_wait(mdl_killer_finished_event);
-
-	os_event_destroy(mdl_killer_stop_event);
-	os_event_destroy(mdl_killer_finished_event);
-}
-
 /*********************************************************************//**
 Function acquires either a backup tables lock, if supported
 by the server, or a global read lock (FLUSH TABLES WITH READ LOCK)
@@ -987,31 +921,6 @@ bool lock_tables(MYSQL *connection)
     return (true);
   }
 
-  if (opt_lock_ddl_per_table)
-  {
-    start_mdl_waiters_killer();
-  }
-
-  if (!opt_lock_wait_timeout && !opt_kill_long_queries_timeout)
-  {
-
-    /* We do first a FLUSH TABLES. If a long update is running, the
-    FLUSH TABLES will wait but will not stall the whole mysqld, and
-    when the long update is done the FLUSH TABLES WITH READ LOCK
-    will start and succeed quickly. So, FLUSH TABLES is to lower
-    the probability of a stage where both mysqldump and most client
-    connections are stalled. Of course, if a second long update
-    starts between the two FLUSHes, we have that bad stall.
-
-    Option lock_wait_timeout serve the same purpose and is not
-    compatible with this trick.
-    */
-
-    msg("Executing FLUSH NO_WRITE_TO_BINLOG TABLES...");
-
-    xb_mysql_query(connection, "FLUSH NO_WRITE_TO_BINLOG TABLES", false);
-  }
-
   if (opt_lock_wait_timeout)
   {
     if (!wait_for_no_updates(connection, opt_lock_wait_timeout,
@@ -1021,7 +930,7 @@ bool lock_tables(MYSQL *connection)
     }
   }
 
-  msg("Executing FLUSH TABLES WITH READ LOCK...");
+  msg("Acquiring BACKUP LOCKS...");
 
   if (opt_kill_long_queries_timeout)
   {
@@ -1030,10 +939,11 @@ bool lock_tables(MYSQL *connection)
 
   if (have_galera_enabled)
   {
-    xb_mysql_query(connection, "SET SESSION wsrep_causal_reads=0", false);
+    xb_mysql_query(connection, "SET SESSION wsrep_sync_wait=0", false);
   }
 
-  xb_mysql_query(connection, "FLUSH TABLES WITH READ LOCK", false, true);
+  xb_mysql_query(connection, "BACKUP STAGE START", true);
+  xb_mysql_query(connection, "BACKUP STAGE BLOCK_COMMIT", true);
   /* Set the maximum supported session value for
   lock_wait_timeout to prevent unnecessary timeouts when the
   global value is changed from the default */
@@ -1041,11 +951,6 @@ bool lock_tables(MYSQL *connection)
     xb_mysql_query(connection, "SET SESSION lock_wait_timeout=31536000",
                    false);
 
-  if (opt_lock_ddl_per_table)
-  {
-    stop_mdl_waiters_killer();
-  }
-
   if (opt_kill_long_queries_timeout)
   {
     stop_query_killer();
@@ -1086,13 +991,8 @@ unlock_all(MYSQL *connection)
 		os_thread_sleep(opt_debug_sleep_before_unlock * 1000);
 	}
 
-	if (binlog_locked) {
-		msg("Executing UNLOCK BINLOG");
-		xb_mysql_query(connection, "UNLOCK BINLOG", false);
-	}
-
-	msg("Executing UNLOCK TABLES");
-	xb_mysql_query(connection, "UNLOCK TABLES", false);
+	msg("Executing BACKUP STAGE END");
+	xb_mysql_query(connection, "BACKUP STAGE END", false);
 
 	msg("All tables unlocked");
 }
@@ -1197,92 +1097,322 @@ cleanup:
 }
 
 
+class Var
+{
+  const char *m_name;
+  char *m_value;
+  /*
+    Disable copying constructors for safety, as the default binary copying
+    which would be wrong. If we ever want them, the m_value
+    member should be copied using an strdup()-alike function.
+  */
+  Var(const Var &); // Disabled
+  Var(Var &);       // Disabled
+public:
+  ~Var()
+  {
+    free(m_value);
+  }
+  Var(const char *name)
+   :m_name(name),
+    m_value(NULL)
+  { }
+  // Init using a SHOW VARIABLES LIKE 'name' query
+  Var(const char *name, MYSQL *mysql)
+   :m_name(name)
+  {
+    char buf[128];
+    my_snprintf(buf, sizeof(buf), "SHOW VARIABLES LIKE '%s'", m_name);
+    m_value= read_mysql_one_value(mysql, buf, 1/*offset*/, 2/*total columns*/);
+  }
+  /*
+    Init by name from a result set.
+    If the variable name is not found in the result set metadata field names,
+    it's value stays untouched.
+  */
+  bool init(MYSQL_RES *mysql_result, MYSQL_ROW row)
+  {
+    MYSQL_FIELD *field= mysql_fetch_fields(mysql_result);
+    for (uint i= 0; i < mysql_num_fields(mysql_result); i++)
+    {
+      if (!strcmp(field[i].name, m_name))
+      {
+        free(m_value); // In case it was initialized earlier
+        m_value= row[i] ? strdup(row[i]) : NULL;
+        return false;
+      }
+    }
+    return true;
+  }
+  void replace(char from, char to)
+  {
+    ut_ad(m_value);
+    for (char *ptr= strchr(m_value, from); ptr; ptr= strchr(ptr, from))
+      *ptr= to;
+  }
+
+  const char *value() const { return m_value; }
+  bool eq_value(const char *str, size_t length) const
+  {
+    return m_value && !strncmp(m_value, str, length) && m_value[length] == '\0';
+  }
+  bool is_null_or_empty() const { return !m_value || !m_value[0]; }
+  bool print(String *to) const
+  {
+    ut_ad(m_value);
+    return to->append(m_value);
+  }
+  bool print_quoted(String *to) const
+  {
+    ut_ad(m_value);
+    return to->append("'") || to->append(m_value) || to->append("'");
+  }
+  bool print_set_global(String *to) const
+  {
+    ut_ad(m_value);
+    return
+      to->append("SET GLOBAL ") ||
+      to->append(m_name) ||
+      to->append(" = '") ||
+      to->append(m_value) ||
+      to->append("';\n");
+  }
+};
+
+
+class Show_slave_status
+{
+  Var m_mariadb_connection_name; // MariaDB: e.g. 'master1'
+  Var m_master;              // e.g. 'localhost'
+  Var m_filename;            // e.g. 'source-bin.000002'
+  Var m_position;            // a number
+  Var m_mysql_gtid_executed; // MySQL56: e.g. single '<UUID>:1-5" or multiline
+                             // '<UUID1>:1-10,\n<UUID2>:1-20\n<UUID3>:1-30'
+  Var m_mariadb_using_gtid;  // MariaDB: 'No','Slave_Pos','Current_Pos'
+
+public:
+
+  Show_slave_status()
+   :m_mariadb_connection_name("Connection_name"),
+    m_master("Master_Host"),
+    m_filename("Relay_Master_Log_File"),
+    m_position("Exec_Master_Log_Pos"),
+    m_mysql_gtid_executed("Executed_Gtid_Set"),
+    m_mariadb_using_gtid("Using_Gtid")
+  { }
+
+  void init(MYSQL_RES *res, MYSQL_ROW row)
+  {
+    m_mariadb_connection_name.init(res, row);
+    m_master.init(res, row);
+    m_filename.init(res, row);
+    m_position.init(res, row);
+    m_mysql_gtid_executed.init(res, row);
+    m_mariadb_using_gtid.init(res, row);
+    // Normalize
+    if (m_mysql_gtid_executed.value())
+      m_mysql_gtid_executed.replace('\n', ' ');
+  }
+
+  static void msg_is_not_slave()
+  {
+    msg("Failed to get master binlog coordinates "
+        "from SHOW SLAVE STATUS.This means that the server is not a "
+        "replication slave. Ignoring the --slave-info option");
+  }
+
+  bool is_mariadb_using_gtid() const
+  {
+    return !m_mariadb_using_gtid.eq_value("No", 2);
+  }
+
+  static bool start_comment_chunk(String *to)
+  {
+    return to->length() ? to->append("; ") : false;
+  }
+
+  bool print_connection_name_if_set(String *to) const
+  {
+    if (!m_mariadb_connection_name.is_null_or_empty())
+      return m_mariadb_connection_name.print_quoted(to) || to->append(' ');
+    return false;
+  }
+
+  bool print_comment_master_identity(String *comment) const
+  {
+    if (comment->append("master "))
+      return true;
+    if (!m_mariadb_connection_name.is_null_or_empty())
+      return m_mariadb_connection_name.print_quoted(comment);
+    return comment->append("''"); // Default not named master
+  }
+
+  bool print_using_master_log_pos(String *sql, String *comment) const
+  {
+    return
+      sql->append("CHANGE MASTER ") ||
+      print_connection_name_if_set(sql) ||
+      sql->append("TO MASTER_LOG_FILE=") || m_filename.print_quoted(sql) ||
+      sql->append(", MASTER_LOG_POS=")   || m_position.print(sql) ||
+      sql->append(";\n") ||
+      print_comment_master_identity(comment) ||
+      comment->append(" filename ")  || m_filename.print_quoted(comment) ||
+      comment->append(" position ")  || m_position.print_quoted(comment);
+  }
+
+  bool print_mysql56(String *sql, String *comment) const
+  {
+    /*
+      SET @@GLOBAL.gtid_purged = '2174B383-5441-11E8-B90A-C80AA9429562:1-1029, '
+                                 '224DA167-0C0C-11E8-8442-00059A3C7B00:1-2695';
+      CHANGE MASTER TO MASTER_AUTO_POSITION=1;
+    */
+    return
+      sql->append("SET GLOBAL gtid_purged=") ||
+      m_mysql_gtid_executed.print_quoted(sql) ||
+      sql->append(";\n") ||
+      sql->append("CHANGE MASTER TO MASTER_AUTO_POSITION=1;\n") ||
+      print_comment_master_identity(comment) ||
+      comment->append(" purge list ") ||
+      m_mysql_gtid_executed.print_quoted(comment);
+  }
+
+  bool print_mariadb10_using_gtid(String *sql, String *comment) const
+  {
+    return
+      sql->append("CHANGE MASTER ") ||
+      print_connection_name_if_set(sql) ||
+      sql->append("TO master_use_gtid = slave_pos;\n") ||
+      print_comment_master_identity(comment) ||
+      comment->append(" master_use_gtid = slave_pos");
+  }
+
+  bool print(String *sql, String *comment, const Var &gtid_slave_pos) const
+  {
+    if (!m_mysql_gtid_executed.is_null_or_empty())
+    {
+      /* MySQL >= 5.6 with GTID enabled */
+      return print_mysql56(sql, comment);
+    }
+
+    if (!gtid_slave_pos.is_null_or_empty() && is_mariadb_using_gtid())
+    {
+      /* MariaDB >= 10.0 with GTID enabled */
+      return print_mariadb10_using_gtid(sql, comment);
+    }
+
+    return print_using_master_log_pos(sql, comment);
+  }
+
+  /*
+    Get master info into strings "sql" and "comment" from a MYSQL_RES.
+    @return false on success
+    @return true on error
+  */
+  static bool get_slave_info(MYSQL_RES *show_slave_info_result,
+                             const Var &gtid_slave_pos,
+                             String *sql, String *comment)
+  {
+    if (!gtid_slave_pos.is_null_or_empty())
+    {
+      // Print gtid_slave_pos if any of the masters really needs it.
+      while (MYSQL_ROW row= mysql_fetch_row(show_slave_info_result))
+      {
+        Show_slave_status status;
+        status.init(show_slave_info_result, row);
+        if (status.is_mariadb_using_gtid())
+        {
+          if (gtid_slave_pos.print_set_global(sql) ||
+              comment->append("gtid_slave_pos ") ||
+              gtid_slave_pos.print_quoted(comment))
+            return true; // Error
+          break;
+        }
+      }
+    }
+
+    // Print the list of masters
+    mysql_data_seek(show_slave_info_result, 0);
+    while (MYSQL_ROW row= mysql_fetch_row(show_slave_info_result))
+    {
+      Show_slave_status status;
+      status.init(show_slave_info_result, row);
+      if (start_comment_chunk(comment) ||
+          status.print(sql, comment, gtid_slave_pos))
+        return true; // Error
+    }
+    return false; // Success
+  }
+
+  /*
+    Get master info into strings "sql" and "comment".
+    @return false on success
+    @return true on error
+  */
+  static bool get_slave_info(MYSQL *mysql, bool show_all_slave_status,
+                             String *sql, String *comment)
+  {
+    bool rc= false; // Success
+    // gtid_slave_pos - MariaDB variable : e.g. "0-1-1" or "1-10-100,2-20-500"
+    Var gtid_slave_pos("gtid_slave_pos", mysql);
+    const char *query= show_all_slave_status ? "SHOW ALL SLAVES STATUS" :
+                                               "SHOW SLAVE STATUS";
+    MYSQL_RES *mysql_result= xb_mysql_query(mysql, query, true);
+    if (!mysql_num_rows(mysql_result))
+    {
+      msg_is_not_slave();
+      // Don't change rc, we still want to continue the backup
+    }
+    else
+    {
+      rc= get_slave_info(mysql_result, gtid_slave_pos, sql, comment);
+    }
+    mysql_free_result(mysql_result);
+    return rc;
+  }
+};
+
+
+
 /*********************************************************************//**
 Retrieves MySQL binlog position of the master server in a replication
 setup and saves it in a file. It also saves it in mysql_slave_position
-variable. */
+variable.
+@returns false on error
+@returns true on success
+*/
 bool
 write_slave_info(MYSQL *connection)
 {
-	char *master = NULL;
-	char *filename = NULL;
-	char *gtid_executed = NULL;
-	char *using_gtid = NULL;
-	char *position = NULL;
-	char *gtid_slave_pos = NULL;
-	char *ptr;
-	bool result = false;
+  String sql, comment;
+  bool show_all_slaves_status= false;
 
-	mysql_variable status[] = {
-		{"Master_Host", &master},
-		{"Relay_Master_Log_File", &filename},
-		{"Exec_Master_Log_Pos", &position},
-		{"Executed_Gtid_Set", &gtid_executed},
-		{"Using_Gtid", &using_gtid},
-		{NULL, NULL}
-	};
+  switch (server_flavor)
+  {
+  case FLAVOR_MARIADB:
+    show_all_slaves_status= mysql_server_version >= 100000;
+    break;
+  case FLAVOR_UNKNOWN:
+  case FLAVOR_MYSQL:
+  case FLAVOR_PERCONA_SERVER:
+    break;
+  }
 
-	mysql_variable variables[] = {
-		{"gtid_slave_pos", &gtid_slave_pos},
-		{NULL, NULL}
-	};
+  if (Show_slave_status::get_slave_info(connection, show_all_slaves_status,
+                                        &sql, &comment))
+    return false; // Error
 
-	read_mysql_variables(connection, "SHOW SLAVE STATUS", status, false);
-	read_mysql_variables(connection, "SHOW VARIABLES", variables, true);
+  if (!sql.length())
+  {
+    /*
+      SHOW [ALL] SLAVE STATUS returned no rows.
+      Don't create the file, but return success to continue the backup.
+    */
+    return true; // Success
+  }
 
-	if (master == NULL || filename == NULL || position == NULL) {
-		msg("Failed to get master binlog coordinates "
-			"from SHOW SLAVE STATUS.This means that the server is not a "
-			"replication slave. Ignoring the --slave-info option");
-		/* we still want to continue the backup */
-		result = true;
-		goto cleanup;
-	}
-
-	/* Print slave status to a file.
-	If GTID mode is used, construct a CHANGE MASTER statement with
-	MASTER_AUTO_POSITION and correct a gtid_purged value. */
-	if (gtid_executed != NULL && *gtid_executed) {
-		/* MySQL >= 5.6 with GTID enabled */
-
-		for (ptr = strchr(gtid_executed, '\n');
-		     ptr;
-		     ptr = strchr(ptr, '\n')) {
-			*ptr = ' ';
-		}
-
-		result = backup_file_printf(XTRABACKUP_SLAVE_INFO,
-			"SET GLOBAL gtid_purged='%s';\n"
-			"CHANGE MASTER TO MASTER_AUTO_POSITION=1\n",
-			gtid_executed);
-
-		ut_a(asprintf(&mysql_slave_position,
-			"master host '%s', purge list '%s'",
-			master, gtid_executed) != -1);
-	} else if (gtid_slave_pos && *gtid_slave_pos &&
-			!(using_gtid && !strncmp(using_gtid, "No", 2))) {
-		/* MariaDB >= 10.0 with GTID enabled */
-		result = backup_file_printf(XTRABACKUP_SLAVE_INFO,
-			"SET GLOBAL gtid_slave_pos = '%s';\n"
-			"CHANGE MASTER TO master_use_gtid = slave_pos\n",
-			gtid_slave_pos);
-		ut_a(asprintf(&mysql_slave_position,
-			"master host '%s', gtid_slave_pos %s",
-			master, gtid_slave_pos) != -1);
-	} else {
-		result = backup_file_printf(XTRABACKUP_SLAVE_INFO,
-			"CHANGE MASTER TO MASTER_LOG_FILE='%s', "
-			"MASTER_LOG_POS=%s\n", filename, position);
-		ut_a(asprintf(&mysql_slave_position,
-			"master host '%s', filename '%s', position '%s'",
-			master, filename, position) != -1);
-	}
-
-cleanup:
-	free_mysql_variables(status);
-	free_mysql_variables(variables);
-
-	return(result);
+  mysql_slave_position= strdup(comment.c_ptr());
+  return backup_file_print_buf(XTRABACKUP_SLAVE_INFO, sql.ptr(), sql.length());
 }
 
 
@@ -1333,27 +1463,29 @@ cleanup:
 }
 
 
+static
+bool
+write_binlog_info(MYSQL *connection, char *log_bin_dir,
+		  MYSQL_RES *mysql_result, my_ulonglong n_rows,
+		  my_ulonglong start);
+
 /*********************************************************************//**
 Flush and copy the current binary log file into the backup,
 if GTID is enabled */
 bool
-write_current_binlog_file(MYSQL *connection)
+write_current_binlog_file(MYSQL *connection, bool write_binlogs)
 {
+	char *log_bin = NULL;
+	char *filename = NULL;
+	char *position = NULL;
 	char *executed_gtid_set = NULL;
 	char *gtid_binlog_state = NULL;
-	char *log_bin_file = NULL;
 	char *log_bin_dir = NULL;
 	bool gtid_exists;
 	bool result = true;
-	char filepath[FN_REFLEN];
 
-	mysql_variable status[] = {
-		{"Executed_Gtid_Set", &executed_gtid_set},
-		{NULL, NULL}
-	};
-
-	mysql_variable status_after_flush[] = {
-		{"File", &log_bin_file},
+	mysql_variable log_bin_var[] = {
+		{"@@GLOBAL.log_bin", &log_bin},
 		{NULL, NULL}
 	};
 
@@ -1363,21 +1495,36 @@ write_current_binlog_file(MYSQL *connection)
 		{NULL, NULL}
 	};
 
+	mysql_variable status[] = {
+		{"File", &filename},
+		{"Position", &position},
+		{"Executed_Gtid_Set", &executed_gtid_set},
+		{NULL, NULL}
+	};
+
+	read_mysql_variables(connection, "SELECT @@GLOBAL.log_bin", log_bin_var, false);
+
+	/* Do not create xtrabackup_binlog_info if binary log is disabled: */
+	if (strncmp(log_bin, "1", 2) != 0) {
+		goto binlog_disabled;
+	}
+
+	lock_binlog_maybe(connection);
+
 	read_mysql_variables(connection, "SHOW MASTER STATUS", status, false);
+
+	/* Do not create xtrabackup_binlog_info if replication
+	has not started yet: */
+	if (filename == NULL || position == NULL) {
+		goto no_replication;
+	}
+
 	read_mysql_variables(connection, "SHOW VARIABLES", vars, true);
 
 	gtid_exists = (executed_gtid_set && *executed_gtid_set)
 			|| (gtid_binlog_state && *gtid_binlog_state);
 
-	if (gtid_exists) {
-		size_t log_bin_dir_length;
-
-		lock_binlog_maybe(connection);
-
-		xb_mysql_query(connection, "FLUSH BINARY LOGS", false);
-
-		read_mysql_variables(connection, "SHOW MASTER STATUS",
-			status_after_flush, false);
+	if (write_binlogs || gtid_exists) {
 
 		if (opt_log_bin != NULL && strchr(opt_log_bin, FN_LIBCHAR)) {
 			/* If log_bin is set, it has priority */
@@ -1387,34 +1534,89 @@ write_current_binlog_file(MYSQL *connection)
 			log_bin_dir = strdup(opt_log_bin);
 		} else if (log_bin_dir == NULL) {
 			/* Default location is MySQL datadir */
-			log_bin_dir = strdup("./");
+			log_bin_dir = static_cast<char*>(malloc(3));
+			ut_a(log_bin_dir);
+			log_bin_dir[0] = '.';
+			log_bin_dir[1] = FN_LIBCHAR;
+			log_bin_dir[2] = 0;
 		}
 
+		size_t log_bin_dir_length;
+
 		dirname_part(log_bin_dir, log_bin_dir, &log_bin_dir_length);
 
 		/* strip final slash if it is not the only path component */
-		if (log_bin_dir_length > 1 &&
-		    log_bin_dir[log_bin_dir_length - 1] == FN_LIBCHAR) {
-			log_bin_dir[log_bin_dir_length - 1] = 0;
+		while (IS_TRAILING_SLASH(log_bin_dir, log_bin_dir_length)) {
+			log_bin_dir_length--;
 		}
+		log_bin_dir[log_bin_dir_length] = 0;
 
-		if (log_bin_dir == NULL || log_bin_file == NULL) {
-			msg("Failed to get master binlog coordinates from "
-				"SHOW MASTER STATUS");
+		if (log_bin_dir == NULL) {
+			msg("Failed to locate binary log files");
 			result = false;
 			goto cleanup;
 		}
 
-		snprintf(filepath, sizeof(filepath), "%s%c%s",
-			 log_bin_dir, FN_LIBCHAR, log_bin_file);
-		result = copy_file(ds_data, filepath, log_bin_file, 0);
+		uint max_binlogs;
+		max_binlogs = opt_max_binlogs;
+		if (max_binlogs == 0) {
+			if (gtid_exists) {
+				max_binlogs = 1;
+			} else {
+				goto cleanup;
+			}
+		}
+
+		xb_mysql_query(connection, "FLUSH BINARY LOGS", false);
+
+		MYSQL_RES *mysql_result;
+
+		mysql_result = xb_mysql_query(connection, "SHOW BINARY LOGS", true);
+
+		ut_ad(mysql_num_fields(mysql_result) >= 2);
+
+		my_ulonglong n_rows;
+		my_ulonglong start;
+
+		n_rows = mysql_num_rows(mysql_result);
+
+		start = 0;
+		if (max_binlogs < n_rows) {
+			start = n_rows - max_binlogs;
+		}
+		if (start) {
+			mysql_data_seek(mysql_result, start);
+		}
+
+		MYSQL_ROW row;
+		while ((row = mysql_fetch_row(mysql_result))) {
+			const char *binlog_name = row[0];
+			char filepath[FN_REFLEN];
+			snprintf(filepath, sizeof(filepath), "%s%c%s",
+				 log_bin_dir, FN_LIBCHAR, binlog_name);
+			if (file_exists(filepath)) {
+				result = copy_file(ds_data, filepath, binlog_name, 0);
+				if (!result) break;
+			}
+		}
+
+		if (result) {
+			write_binlog_info(connection, log_bin_dir,
+					  mysql_result, n_rows, start);
+		}
+
+		mysql_free_result(mysql_result);
 	}
 
 cleanup:
-	free_mysql_variables(status_after_flush);
-	free_mysql_variables(status);
 	free_mysql_variables(vars);
 
+no_replication:
+	free_mysql_variables(status);
+
+binlog_disabled:
+	free_mysql_variables(log_bin_var);
+
 	return(result);
 }
 
@@ -1422,8 +1624,11 @@ cleanup:
 /*********************************************************************//**
 Retrieves MySQL binlog position and
 saves it in a file. It also prints it to stdout. */
+static
 bool
-write_binlog_info(MYSQL *connection)
+write_binlog_info(MYSQL *connection, char *log_bin_dir,
+		  MYSQL_RES *mysql_result, my_ulonglong n_rows,
+		  my_ulonglong start)
 {
 	char *filename = NULL;
 	char *position = NULL;
@@ -1431,9 +1636,13 @@ write_binlog_info(MYSQL *connection)
 	char *gtid_current_pos = NULL;
 	char *gtid_executed = NULL;
 	char *gtid = NULL;
-	bool result;
+	char *buffer;
+	char *buf;
+	size_t total;
+	bool result = true;
 	bool mysql_gtid;
 	bool mariadb_gtid;
+	bool with_gtid;
 
 	mysql_variable status[] = {
 		{"File", &filename},
@@ -1451,39 +1660,106 @@ write_binlog_info(MYSQL *connection)
 	read_mysql_variables(connection, "SHOW MASTER STATUS", status, false);
 	read_mysql_variables(connection, "SHOW VARIABLES", vars, true);
 
-	if (filename == NULL || position == NULL) {
-		/* Do not create xtrabackup_binlog_info if binary
-		log is disabled */
-		result = true;
-		goto cleanup;
-	}
+	mysql_gtid = gtid_mode && (strcmp(gtid_mode, "ON") == 0);
+	mariadb_gtid = gtid_current_pos && *gtid_current_pos;
 
-	mysql_gtid = ((gtid_mode != NULL) && (strcmp(gtid_mode, "ON") == 0));
-	mariadb_gtid = (gtid_current_pos != NULL);
+	gtid = (gtid_executed && *gtid_executed) ? gtid_executed : gtid_current_pos;
 
-	gtid = (gtid_executed != NULL ? gtid_executed : gtid_current_pos);
-
-	if (mariadb_gtid || mysql_gtid) {
+	with_gtid = mariadb_gtid || mysql_gtid;
+	if (with_gtid) {
 		ut_a(asprintf(&mysql_binlog_position,
 			"filename '%s', position '%s', "
 			"GTID of the last change '%s'",
 			filename, position, gtid) != -1);
-		result = backup_file_printf(XTRABACKUP_BINLOG_INFO,
-					    "%s\t%s\t%s\n", filename, position,
-					    gtid);
 	} else {
 		ut_a(asprintf(&mysql_binlog_position,
 			"filename '%s', position '%s'",
 			filename, position) != -1);
-		result = backup_file_printf(XTRABACKUP_BINLOG_INFO,
-					    "%s\t%s\n", filename, position);
 	}
 
+	mysql_data_seek(mysql_result, start);
+
+	MYSQL_ROW row;
+	my_ulonglong current;
+
+	total = 1;
+	current = start;
+	while ((row = mysql_fetch_row(mysql_result))) {
+		const char *binlog_name = row[0];
+		/* The position in the current binlog is taken from
+		the global variable, but for the previous ones it is
+		determined by their length: */
+		const char *binlog_pos =
+			++current == n_rows ? position : row[1];
+		total += strlen(binlog_name) + strlen(binlog_pos) + 2;
+		if (with_gtid && current != n_rows) {
+			/* Add the "\t[]" length to the buffer size: */
+			total += 3;
+		}
+	}
+	/* For the last of the binray log files, also add
+	the length of the GTID (+ one character for '\t'): */
+	if (with_gtid) {
+		total += strlen(gtid) + 1;
+	}
+
+	buffer = static_cast<char*>(malloc(total));
+	if (!buffer) {
+		msg("Failed to allocate memory for temporary buffer");
+		result = false;
+		goto cleanup;
+	}
+
+	mysql_data_seek(mysql_result, start);
+
+	buf = buffer;
+	current = start;
+	while ((row = mysql_fetch_row(mysql_result))) {
+		const char *binlog_name = row[0];
+		char filepath[FN_REFLEN];
+		snprintf(filepath, sizeof(filepath), "%s%c%s",
+			 log_bin_dir, FN_LIBCHAR, binlog_name);
+		current++;
+		if (file_exists(filepath)) {
+			/* The position in the current binlog is taken from
+			the global variable, but for the previous ones it is
+			determined by their length: */
+			char *binlog_pos =
+				current == n_rows ? position : row[1];
+			int bytes;
+			if (with_gtid) {
+				bytes = snprintf(buf, total, "%s\t%s\t%s\n",
+						 binlog_name, binlog_pos,
+						 current == n_rows ? gtid : "[]");
+			} else {
+				bytes = snprintf(buf, total, "%s\t%s\n",
+						 binlog_name, binlog_pos);
+			}
+			if (bytes <= 0) {
+				goto buffer_overflow;
+			}
+			buf += bytes;
+			total -= bytes;
+		}
+	}
+
+	if (buf != buffer) {
+		result = backup_file_printf(XTRABACKUP_BINLOG_INFO, "%s", buffer);
+	}
+
+cleanup2:
+	free(buffer);
+
 cleanup:
-	free_mysql_variables(status);
 	free_mysql_variables(vars);
+	free_mysql_variables(status);
 
 	return(result);
+
+buffer_overflow:
+	msg("Internal error: buffer overflow in the write_binlog_info()");
+	result = false;
+	goto cleanup2;
 }
 
 struct escape_and_quote
@@ -1503,7 +1779,7 @@ operator<<(std::ostream& s, const escape_and_quote& eq)
 	s << '\'';
 	size_t len = strlen(eq.str);
 	char* escaped = (char *)alloca(2 * len + 1);
-	len = mysql_real_escape_string(eq.mysql, escaped, eq.str, len);
+	len = mysql_real_escape_string(eq.mysql, escaped, eq.str, (ulong)len);
 	s << std::string(escaped, len);
 	s << '\'';
 	return s;
@@ -1811,7 +2087,6 @@ backup_cleanup()
 {
 	free(mysql_slave_position);
 	free(mysql_binlog_position);
-	free(buffer_pool_filename);
 
 	if (mysql_connection) {
 		mysql_close(mysql_connection);
