@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2016 MariaDB Corporation
+  Copyright (c) 2016, 2021, MariaDB Corporation.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
   All methods pertaining to temporary tables.
 */
 
+#include "mariadb.h"
 #include "sql_acl.h"                            /* TMP_TABLE_ACLS */
 #include "sql_base.h"                           /* tdc_create_key */
 #include "lock.h"                               /* mysql_lock_remove */
@@ -48,32 +49,30 @@ bool THD::has_thd_temporary_tables()
 /**
   Create a temporary table, open it and return the TABLE handle.
 
-  @param hton [IN]                    Handlerton
   @param frm  [IN]                    Binary frm image
   @param path [IN]                    File path (without extension)
   @param db   [IN]                    Schema name
   @param table_name [IN]              Table name
-  @param open_in_engine [IN]          Whether open table in SE
-
 
   @return Success                     A pointer to table object
           Failure                     NULL
 */
-TABLE *THD::create_and_open_tmp_table(handlerton *hton,
-                                      LEX_CUSTRING *frm,
+TABLE *THD::create_and_open_tmp_table(LEX_CUSTRING *frm,
                                       const char *path,
                                       const char *db,
                                       const char *table_name,
-                                      bool open_in_engine)
+                                      bool open_internal_tables)
 {
   DBUG_ENTER("THD::create_and_open_tmp_table");
 
   TMP_TABLE_SHARE *share;
   TABLE *table= NULL;
 
-  if ((share= create_temporary_table(hton, frm, path, db, table_name)))
+  if ((share= create_temporary_table(frm, path, db, table_name)))
   {
-    table= open_temporary_table(share, table_name, open_in_engine);
+    open_options|= HA_OPEN_FOR_CREATE;
+    table= open_temporary_table(share, table_name);
+    open_options&= ~HA_OPEN_FOR_CREATE;
 
     /*
       Failed to open a temporary table instance. As we are not passing
@@ -87,6 +86,15 @@ TABLE *THD::create_and_open_tmp_table(handlerton *hton,
 
       /* Free the TMP_TABLE_SHARE. */
       free_tmp_table_share(share, false);
+      DBUG_RETURN(0);
+    }
+
+    /* Open any related tables */
+    if (open_internal_tables && table->internal_tables &&
+        open_and_lock_internal_tables(table, true))
+    {
+      drop_temporary_table(table, NULL, false);
+      DBUG_RETURN(0);
     }
   }
 
@@ -248,7 +256,7 @@ TMP_TABLE_SHARE *THD::find_tmp_table_share(const TABLE_LIST *tl)
   @return Success                     A pointer to table share object
           Failure                     NULL
 */
-TMP_TABLE_SHARE *THD::find_tmp_table_share(const char *key, uint key_length)
+TMP_TABLE_SHARE *THD::find_tmp_table_share(const char *key, size_t key_length)
 {
   DBUG_ENTER("THD::find_tmp_table_share");
 
@@ -313,7 +321,7 @@ TMP_TABLE_SHARE *THD::find_tmp_table_share(const char *key, uint key_length)
 bool THD::open_temporary_table(TABLE_LIST *tl)
 {
   DBUG_ENTER("THD::open_temporary_table");
-  DBUG_PRINT("enter", ("table: '%s'.'%s'", tl->db, tl->table_name));
+  DBUG_PRINT("enter", ("table: '%s'.'%s'", tl->db.str, tl->table_name.str));
 
   TMP_TABLE_SHARE *share;
   TABLE *table= NULL;
@@ -338,7 +346,7 @@ bool THD::open_temporary_table(TABLE_LIST *tl)
     DBUG_RETURN(false);
   }
 
-  if (!tl->db)
+  if (!tl->db.str)
   {
     DBUG_PRINT("info",
                ("Table reference to a temporary table must have database set"));
@@ -375,7 +383,7 @@ bool THD::open_temporary_table(TABLE_LIST *tl)
   */
   if (!table && (share= find_tmp_table_share(tl)))
   {
-    table= open_temporary_table(share, tl->get_table_name(), true);
+    table= open_temporary_table(share, tl->get_table_name());
     /*
        Temporary tables are not safe for parallel replication. They were
        designed to be visible to one thread only, so have no table locking.
@@ -399,7 +407,7 @@ bool THD::open_temporary_table(TABLE_LIST *tl)
     if (tl->open_type == OT_TEMPORARY_ONLY &&
         tl->open_strategy == TABLE_LIST::OPEN_NORMAL)
     {
-      my_error(ER_NO_SUCH_TABLE, MYF(0), tl->db, tl->table_name);
+      my_error(ER_NO_SUCH_TABLE, MYF(0), tl->db.str, tl->table_name.str);
       DBUG_RETURN(true);
     }
     DBUG_RETURN(false);
@@ -553,25 +561,21 @@ bool THD::close_temporary_tables()
           true                        Error
 */
 bool THD::rename_temporary_table(TABLE *table,
-                                 const char *db,
-                                 const char *table_name)
+                                 const LEX_CSTRING *db,
+                                 const LEX_CSTRING *table_name)
 {
-  DBUG_ENTER("THD::rename_temporary_table");
-
   char *key;
   uint key_length;
-
   TABLE_SHARE *share= table->s;
+  DBUG_ENTER("THD::rename_temporary_table");
 
   if (!(key= (char *) alloc_root(&share->mem_root, MAX_DBKEY_LENGTH)))
-  {
     DBUG_RETURN(true);
-  }
 
   /*
     Temporary tables are renamed by simply changing their table definition key.
   */
-  key_length= create_tmp_table_def_key(key, db, table_name);
+  key_length= create_tmp_table_def_key(key, db->str, table_name->str);
   share->set_table_cache_key(key, key_length);
 
   DBUG_RETURN(false);
@@ -689,19 +693,20 @@ bool THD::rm_temporary_table(handlerton *base, const char *path)
   char frm_path[FN_REFLEN + 1];
 
   strxnmov(frm_path, sizeof(frm_path) - 1, path, reg_ext, NullS);
-  if (mysql_file_delete(key_file_frm, frm_path, MYF(0)))
-  {
-    error= true;
-  }
-  file= get_new_handler((TABLE_SHARE*) 0, current_thd->mem_root, base);
+
+  file= get_new_handler((TABLE_SHARE*) 0, mem_root, base);
   if (file && file->ha_delete_table(path))
   {
     error= true;
     sql_print_warning("Could not remove temporary table: '%s', error: %d",
                       path, my_errno);
   }
-
   delete file;
+
+  if (mysql_file_delete(key_file_frm, frm_path, MYF(0)))
+  {
+    error= true;
+  }
   DBUG_RETURN(error);
 }
 
@@ -745,6 +750,8 @@ void THD::mark_tmp_tables_as_free_for_reuse()
     {
       if ((table->query_id == query_id) && !table->open_by_handler)
       {
+        if (table->update_handler)
+          table->delete_update_handler();
         mark_tmp_table_as_free_for_reuse(table);
       }
     }
@@ -869,10 +876,12 @@ void THD::restore_tmp_table_share(TMP_TABLE_SHARE *share)
 bool THD::has_temporary_tables()
 {
   DBUG_ENTER("THD::has_temporary_tables");
-  bool result= (rgi_slave
-                ? (rgi_slave->rli->save_temporary_tables &&
-                   !rgi_slave->rli->save_temporary_tables->is_empty())
-                : (has_thd_temporary_tables()));
+  bool result=
+#ifdef HAVE_REPLICATION
+    rgi_slave ? (rgi_slave->rli->save_temporary_tables &&
+                 !rgi_slave->rli->save_temporary_tables->is_empty()) :
+#endif
+    has_thd_temporary_tables();
   DBUG_RETURN(result);
 }
 
@@ -917,7 +926,6 @@ uint THD::create_tmp_table_def_key(char *key, const char *db,
 /**
   Create a temporary table.
 
-  @param hton [IN]                    Handlerton
   @param frm  [IN]                    Binary frm image
   @param path [IN]                    File path (without extension)
   @param db   [IN]                    Schema name
@@ -926,8 +934,7 @@ uint THD::create_tmp_table_def_key(char *key, const char *db,
   @return Success                     A pointer to table share object
           Failure                     NULL
 */
-TMP_TABLE_SHARE *THD::create_temporary_table(handlerton *hton,
-                                             LEX_CUSTRING *frm,
+TMP_TABLE_SHARE *THD::create_temporary_table(LEX_CUSTRING *frm,
                                              const char *path,
                                              const char *db,
                                              const char *table_name)
@@ -965,8 +972,6 @@ TMP_TABLE_SHARE *THD::create_temporary_table(handlerton *hton,
   init_tmp_table_share(this, share, saved_key_cache, key_length,
                        strend(saved_key_cache) + 1, tmp_path);
 
-  share->db_plugin= ha_lock_engine(this, hton);
-
   /*
     Prefer using frm image over file. The image might not be available in
     ALTER TABLE, when the discovering engine took over the ownership (see
@@ -1093,28 +1098,29 @@ TABLE *THD::find_temporary_table(const char *key, uint key_length,
 
   @param share [IN]                   Table share
   @param alias [IN]                   Table alias
-  @param open_in_engine [IN]          Whether open table in SE
 
   @return Success                     A pointer to table object
           Failure                     NULL
 */
 TABLE *THD::open_temporary_table(TMP_TABLE_SHARE *share,
-                                 const char *alias,
-                                 bool open_in_engine)
+                                 const char *alias_arg)
 {
+  TABLE *table;
+  LEX_CSTRING alias= {alias_arg, strlen(alias_arg) };
   DBUG_ENTER("THD::open_temporary_table");
 
-  TABLE *table;
 
   if (!(table= (TABLE *) my_malloc(sizeof(TABLE), MYF(MY_WME))))
   {
     DBUG_RETURN(NULL);                          /* Out of memory */
   }
 
-  if (open_table_from_share(this, share, alias,
-                            open_in_engine ? (uint)HA_OPEN_KEYFILE : 0,
-                            EXTRA_RECORD, open_options | ha_open_options, table,
-                            open_in_engine ? false : true))
+  if (open_table_from_share(this, share, &alias,
+                            (uint) HA_OPEN_KEYFILE,
+                            EXTRA_RECORD,
+                            (ha_open_options |
+                             (open_options & HA_OPEN_FOR_CREATE)),
+                            table, false))
   {
     my_free(table);
     DBUG_RETURN(NULL);
@@ -1122,8 +1128,9 @@ TABLE *THD::open_temporary_table(TMP_TABLE_SHARE *share,
 
   table->reginfo.lock_type= TL_WRITE;           /* Simulate locked */
   table->grant.privilege= TMP_TABLE_ACLS;
-  share->tmp_table= (table->file->has_transactions() ?
+  share->tmp_table= (table->file->has_transaction_manager() ?
                      TRANSACTIONAL_TMP_TABLE : NON_TRANSACTIONAL_TMP_TABLE);
+  share->not_usable_by_query_cache= 1;
 
   table->pos_in_table_list= 0;
   table->query_id= query_id;
@@ -1133,11 +1140,10 @@ TABLE *THD::open_temporary_table(TMP_TABLE_SHARE *share,
 
   /* Increment Slave_open_temp_table_definitions status variable count. */
   if (rgi_slave)
-  {
-    thread_safe_increment32(&slave_open_temp_tables);
-  }
+    slave_open_temp_tables++;
 
-  DBUG_PRINT("tmptable", ("Opened table: '%s'.'%s'%p", table->s->db.str,
+  DBUG_PRINT("tmptable", ("Opened table: '%s'.'%s  table: %p",
+                          table->s->db.str,
                           table->s->table_name.str, table));
   DBUG_RETURN(table);
 }
@@ -1240,7 +1246,7 @@ void THD::close_temporary_table(TABLE *table)
     /* Natural invariant of temporary_tables */
     DBUG_ASSERT(slave_open_temp_tables || !temporary_tables);
     /* Decrement Slave_open_temp_table_definitions status variable count. */
-    thread_safe_decrement32(&slave_open_temp_tables);
+    slave_open_temp_tables--;
   }
 
   DBUG_VOID_RETURN;
@@ -1373,15 +1379,15 @@ bool THD::log_events_and_free_tmp_shares()
           /*
              We are going to add ` around the table names and possible more
              due to special characters.
-           */
-          append_identifier(this, &s_query, share->table_name.str,
-              share->table_name.length);
+          */
+          append_identifier(this, &s_query, &share->table_name);
           s_query.append(',');
         }
         rm_temporary_table(share->db_type(), share->path.str);
         free_table_share(share);
         my_free(share);
       }
+
       if (at_least_one_create_logged)
       {
         clear_error();
@@ -1398,7 +1404,8 @@ bool THD::log_events_and_free_tmp_shares()
 
         get_stmt_da()->set_overwrite_status(true);
         transaction.stmt.mark_dropped_temp_table();
-        if ((error= (mysql_bin_log.write(&qinfo) || error)))
+        bool error2= mysql_bin_log.write(&qinfo);
+        if (unlikely(error|= error2))
         {
           /*
              If we're here following THD::cleanup, thence the connection
@@ -1411,7 +1418,7 @@ bool THD::log_events_and_free_tmp_shares()
              that THD::close_tables failed. (Actually, the SQL
              thread only calls THD::close_tables while applying
              old Start_log_event_v3 events.)
-           */
+          */
           sql_print_error("Failed to write the DROP statement for "
               "temporary tables to binary log");
         }
@@ -1504,12 +1511,14 @@ bool THD::lock_temporary_tables()
     DBUG_RETURN(false);
   }
 
+#ifdef HAVE_REPLICATION
   if (rgi_slave)
   {
     mysql_mutex_lock(&rgi_slave->rli->data_lock);
     temporary_tables= rgi_slave->rli->save_temporary_tables;
     m_tmp_tables_locked= true;
   }
+#endif
 
   DBUG_RETURN(m_tmp_tables_locked);
 }
@@ -1530,6 +1539,7 @@ void THD::unlock_temporary_tables()
     DBUG_VOID_RETURN;
   }
 
+#ifdef HAVE_REPLICATION
   if (rgi_slave)
   {
     rgi_slave->rli->save_temporary_tables= temporary_tables;
@@ -1537,6 +1547,7 @@ void THD::unlock_temporary_tables()
     mysql_mutex_unlock(&rgi_slave->rli->data_lock);
     m_tmp_tables_locked= false;
   }
+#endif
 
   DBUG_VOID_RETURN;
 }
